import json
import os
import random
from pathlib import Path
from urllib.parse import urlparse

import boto3
import geopandas as gp
import planetary_computer as pc
import pyarrow as pa
import pystac_client
import rasterio
from geoarrow.pyarrow import io
from rasterio.io import MemoryFile

from stacchip.indexer import NoDataMaskChipIndexer

STAC_API = "https://planetarycomputer.microsoft.com/api/stac/v1"
S1_ASSETS = [
    "vv",
    "vh",
]
PLATFORM_NAME = "sentinel-1-rtc"
quartals = [
    "{year}-01-01/{year}-03-31",
    "{year}-04-01/{year}-06-30",
    "{year}-07-01/{year}-09-30",
    "{year}-10-01/{year}-12-31",
]


def process_mgrs_tile(index: int, mgrs_source: str, bucket: str) -> None:
    # Prepare resources for the job
    s3 = boto3.resource("s3")
    data = gp.read_file(mgrs_source)
    row = data.iloc[index]
    catalog = pystac_client.Client.open(STAC_API, modifier=pc.sign_inplace)
    print("MGRS", row["name"])
    random.seed(index)
    for year in random.sample(range(2018, 2024), 1):
        print(f"Year {year}")
        for quartal in random.sample(quartals, 1):
            print(f"Quartal {quartal.format(year=year)}")
            items = catalog.search(
                max_items=1,
                filter_lang="cql2-json",
                filter={
                    "op": "and",
                    "args": [
                        # {
                        #     "op": "s_intersects",
                        #     "args": [
                        #         {"property": "geometry"},
                        #         row.geometry.__geo_interface__,
                        #     ],
                        # },
                        {
                            "op": "anyinteracts",
                            "args": [
                                {"property": "datetime"},
                                quartal.format(year=year),
                            ],
                        },
                        {
                            "op": "=",
                            "args": [{"property": "collection"}, "sentinel-1-rtc"],
                        },
                    ],
                },
            )
            item = items.item_collection()[0]

            nodata_mask = None
            for key in list(item.assets.keys()):
                if key not in S1_ASSETS:
                    del item.assets[key]
                else:
                    url = item.assets[key].href
                    with rasterio.open(url) as rst:
                        data = rst.read()
                        meta = rst.meta.copy()
                        if nodata_mask is None:
                            nodata_mask = data[0] == rst.nodata

                    with MemoryFile() as memfile:
                        with memfile.open(**meta, compress="deflate") as dst:
                            dst.write(data)

                        memfile.seek(0)

                        s3_bucket = s3.Bucket(name=bucket)
                        new_key = (
                            f"{PLATFORM_NAME}/{item.id}/{Path(urlparse(url).path).name}"
                        )
                        print(f"Copying {urlparse(url).path}")
                        s3_bucket.put_object(
                            Key=new_key,
                            Body=memfile.read(),
                        )

                    item.assets[key].href = f"s3://{bucket}/{new_key}"

            # Convert Dictionary to JSON String
            data_string = json.dumps(item.to_dict())

            # Upload JSON String to an S3 Object
            s3_bucket = s3.Bucket(name=bucket)
            s3_bucket.put_object(
                Key=f"{PLATFORM_NAME}/{item.id}/stac_item.json",
                Body=data_string,
            )
            indexer = NoDataMaskChipIndexer(item, nodata_mask=nodata_mask)
            index = indexer.create_index()

            writer = pa.BufferOutputStream()
            io.write_geoparquet_table(index, writer)
            body = bytes(writer.getvalue())
            # Centralize the index files to make combining them easier later on
            s3_bucket.put_object(
                Body=body,
                Key=f"index/{PLATFORM_NAME}/{item.id}/index_{item.id}.parquet",
            )


def process() -> None:

    if "AWS_BATCH_JOB_ARRAY_INDEX" not in os.environ:
        raise ValueError("AWS_BATCH_JOB_ARRAY_INDEX env var not set")
    if "STACCHIP_MGRS_SOURCE" not in os.environ:
        raise ValueError("STACCHIP_MGRS_SOURCE env var not set")
    if "STACCHIP_BUCKET" not in os.environ:
        raise ValueError("STACCHIP_BUCKET env var not set")

    index = int(os.environ["AWS_BATCH_JOB_ARRAY_INDEX"])
    mgrs_source = os.environ["STACCHIP_MGRS_SOURCE"]
    bucket = os.environ["STACCHIP_BUCKET"]

    process_mgrs_tile(index, mgrs_source, bucket)



import pystac_client
from stacchip.indexer import Sentinel2Indexer
from stacchip.chipper import Chipper
import os

catalog = pystac_client.Client.open("https://earth-search.aws.element84.com/v1")

os.environ["CPL_TMPDIR"] = "/tmp"
os.environ["GDAL_CACHEMAX"] = "75%"
os.environ["GDAL_INGESTED_BYTES_AT_OPEN"] = "32768"
os.environ["GDAL_DISABLE_READDIR_ON_OPEN"] = "EMPTY_DIR"
os.environ["GDAL_HTTP_MERGE_CONSECUTIVE_RANGES"] = "YES"
os.environ["GDAL_HTTP_MULTIPLEX"] = "YES"
os.environ["GDAL_HTTP_VERSION"] = "2"
os.environ["PYTHONWARNINGS"] = "ignore"
os.environ["VSI_CACHE"] = "TRUE"


items = catalog.search(collections=["sentinel-2-l2a"], max_items=1)
item = items.item_collection()[0]
dat = list(item.assets.keys())
for k in dat:
    if k not in ["red", "green", "blue", "nir"]:
        del item.assets[k]
indexer = Sentinel2Indexer(item)
# index = .create_index()
chip = Chipper(item_id=item.id, indexer=indexer, platform="", bucket="example").chip(
    x=10, y=10
)

from pyarrow import dataset as da

from geoarrow.pyarrow import io

import geoarrow.pyarrow as ga

table = io.read_geoparquet_table(
    "/home/tam/Desktop/clay-v1-data-combined-index.parquet"
)
geoms = table.column("geometry")
pdgeoms = ga.to_geopandas(geoms)

centroids = pdgeoms.centroid


data = da.dataset(
    "/home/tam/Desktop/clay-v1-data-combined-index.parquet", format="parquet"
)
table = data.to_table(
    columns=["chipid", "platform", "item", "date", "chip_index_x", "chip_index_y"]
)



import json
from pathlib import Path

import boto3
from pystac import Item

from stacchip.indexer import ChipIndexer


def load_indexer_s3(bucket: str, platform: str, item_id: str) -> ChipIndexer:
    """
    Load stacchip index table from a remote location
    """
    s3 = boto3.resource("s3")
    s3_bucket = s3.Bucket(name=bucket)
    content_object = s3_bucket.Object(f"{platform}/{item_id}/stac_item.json")
    file_content = content_object.get()["Body"].read().decode("utf-8")
    json_content = json.loads(file_content)
    item = Item.from_dict(json_content)

    return ChipIndexer(item)


def load_indexer_local(mountpath: Path, platform: str, item_id: str) -> ChipIndexer:
    """
    Load stacchip index table from local file
    """
    item = Item.from_file(mountpath / Path(f"{platform}/{item_id}/stac_item.json"))
    return ChipIndexer(item)



import io
import os
from multiprocessing import Pool

import boto3
import numpy as np

from stacchip.processors.prechip import (
    LINZ_BANDS,
    LS_BANDS,
    NAIP_BANDS,
    S1_BANDS,
    S2_BANDS,
)


def get_stats_keys(key):
    print(f"Processing {key}")
    if "sentinel-1-rtc" in key:
        nodata = -32768
    else:
        nodata = 0

    s3_session = boto3.resource("s3")
    obj = s3_session.Object("clay-v1-data-cubes", key)
    body = obj.get()["Body"].read()
    with io.BytesIO(body) as f:
        f.seek(0)
        data = np.load(f)["pixels"]

    data = data.astype("float64").swapaxes(0, 1)

    data = np.ma.array(data, mask=data == nodata)

    pixel_count = np.ma.count(data, axis=(1, 2, 3))
    pixel_sum = np.ma.sum(data, axis=(1, 2, 3))
    pixel_sqr = np.ma.sum(np.ma.power(data, 2), axis=(1, 2, 3))

    return pixel_count, pixel_sum, pixel_sqr


def process():
    if "STACCHIP_PLATFORM" not in os.environ:
        raise ValueError("STACCHIP_PLATFORM env var not set")
    pool_size = int(os.environ.get("STACCHIP_POOL_SIZE", 4))
    max_cubes = int(os.environ.get("STACCHIP_MAX_CUBES", 4))

    platform = os.environ.get("STACCHIP_PLATFORM")
    if platform == "naip":
        bands = NAIP_BANDS
    elif platform == "linz":
        bands = LINZ_BANDS
    elif platform == "sentinel-2-l2a":
        bands = S2_BANDS
    elif platform in ["landsat-c2l2-sr", "landsat-c2l1"]:
        bands = LS_BANDS
    elif platform == "sentinel-1-rtc":
        bands = S1_BANDS
    else:
        raise ValueError(f"Platform {platform} not found")

    client = boto3.client("s3")
    paginator = client.get_paginator("list_objects_v2")
    page_iterator = paginator.paginate(
        Bucket="clay-v1-data-cubes", Prefix=f"mode_v1_chipper_v2/{platform}"
    )

    band_count = len(bands)
    pixel_count = np.zeros(band_count)
    pixel_sum = np.zeros(band_count)
    pixel_sqr = np.zeros(band_count)

    counter = 0
    all_keys = []
    for page in page_iterator:
        keys = [dat["Key"] for dat in page["Contents"]]
        for key in keys:
            counter += 1
            all_keys.append(key)
            if counter == max_cubes:
                break
        if counter == max_cubes:
            break

    with Pool(pool_size) as pl:
        result = pl.map(get_stats_keys, all_keys)

    for dat in result:
        pixel_count = np.add(pixel_count, dat[0])
        pixel_sum = np.add(pixel_sum, dat[1])
        pixel_sqr = np.add(pixel_sqr, dat[2])

    # https://stackoverflow.com/questions/1174984/how-to-efficiently-calculate-a-running-standard-deviation
    mean = pixel_sum / pixel_count
    stdev = np.sqrt((pixel_sqr / pixel_count) - (mean * mean))

    print("-- Mean by band")
    for band, val in zip(bands, mean):
        print(f"{band}: {val}")

    print("-- Std by band")
    for band, val in zip(bands, stdev):
        print(f"{band}: {val}")



import json
import os
import random
import tempfile
from pathlib import Path

import boto3
import pyarrow as pa
import rasterio
from dateutil import parser
from geoarrow.pyarrow import io
from pystac import Item
from rasterio.enums import Resampling
from rio_stac import create_stac_item

from stacchip.indexer import NoDataMaskChipIndexer

PLATFORM_NAME = "linz"

TARGET_RESOLUTION = 0.3

nz_prefixes = [
    "auckland/auckland_2022_0.075m/",
    "wellington/wellington_2021_0.075m/",
    "wellington/wellington_2021_0.3m/",
    "bay-of-plenty/bay-of-plenty_2023_0.1m/",
    "bay-of-plenty/tauranga_2022_0.1m/",
    "bay-of-plenty/tauranga-winter_2022_0.1m/",
    "canterbury/canterbury_2023_0.3m/",
    "canterbury/ashburton_2023_0.1m/",
    "canterbury/hurunui_2023_0.075m/",
    "canterbury/timaru_2022-2023_0.1m/",
    "canterbury/selwyn_2022-2023_0.075m/",
    "gisborne/gisborne_2023_0.075m/",
    "hawkes-bay/hawkes-bay_2022_0.05m/",
    "hawkes-bay/napier_2017-2018_0.05m/",
    "hawkes-bay/wairoa_2014-2015_0.1m/",
    "manawatu-whanganui/manawatu-whanganui_2021-2022_0.3m/",
    "manawatu-whanganui/palmerston-north_2022_0.125m/",
    "manawatu-whanganui/rangitikei_2021_0.125m/",
    "manawatu-whanganui/tararua_2024_0.1m/",
    "manawatu-whanganui/whanganui_2022_0.075m/",
    "marlborough/marlborough_2023_0.075m/",
    "nelson/nelson_2022_0.075m/",
    "northland/northland_2016_0.1m/",
    "otago/queenstown_2021_0.1m/",
    "otago/otago_2018_0.1m/",
    "otago/dunedin_2018-2019_0.1m/",
    "southland/southland_2023_0.1m/",
    "southland/invercargill_2022_0.05m/",
    "taranaki/taranaki_2022_0.05m/",
    "taranaki/new-plymouth_2017_0.1m/",
    "tasman/tasman_2023_0.075m/",
    "waikato/hamilton_2023_0.05m/",
    "waikato/otorohanga_2021_0.1m/",
    "waikato/taupo_2023_0.075m/",
    "waikato/thames-coromandel_2021_0.1m/",
    "waikato/waikato_2021_0.1m/",
    "waikato/waipa_2021_0.1m/",
    "west-coast/buller_2020_0.2m/",
    "west-coast/west-coast_2016_0.1m/",
]


def get_linz_tiffs(prefix) -> list:

    s3_resource = boto3.resource("s3")
    s3_bucket = s3_resource.Bucket(name="nz-imagery")

    files = []
    s3_object_iterator = s3_bucket.objects.filter(Prefix=prefix)

    for s3_object in s3_object_iterator:
        if s3_object.key.endswith(".tiff"):
            files.append(s3_object.key)

    # Sample a percentage of all scenes
    sample_size = max(min(int(len(files) / 2), 2000), 10)
    print(f"Found {len(files)} scenes for {prefix}, keeping {sample_size}")
    random.seed(42)
    return random.sample(files, sample_size)


def get_original_item(key: str) -> Item:
    s3_resource = boto3.resource("s3")
    content_object = s3_resource.Object(
        "nz-imagery", key.replace(".tiff", "") + ".json"
    )
    file_content = content_object.get()["Body"].read().decode("utf-8")
    json_content = json.loads(file_content)
    return Item.from_dict(json_content)


def process_linz_tile(index, bucket):

    tiffs = get_linz_tiffs(nz_prefixes[index])

    for key in tiffs:
        print(f"Working on {key}")

        href = f"s3://nz-imagery/{key}"

        original_item = get_original_item(key)

        # For now, resample so we have a constant gsd for all images
        with rasterio.open(href) as dataset:

            gsd = abs(dataset.transform[0])

            upscale_factor = gsd / TARGET_RESOLUTION

            data = dataset.read(
                out_shape=(
                    dataset.count,
                    int(dataset.height * upscale_factor),
                    int(dataset.width * upscale_factor),
                ),
                resampling=Resampling.bilinear,
            )

            # Drop alpha band if present
            data = data[:3]

            # scale image transform
            transform = dataset.transform * dataset.transform.scale(
                (dataset.width / data.shape[-1]), (dataset.height / data.shape[-2])
            )

            new_key = f"{PLATFORM_NAME}/{original_item.id}/{Path(href).name}"
            new_href = f"s3://{bucket}/{new_key}"

            meta = dataset.meta.copy()
            meta["transform"] = transform
            meta["width"] = data.shape[2]
            meta["height"] = data.shape[1]
            meta["compress"] = "deflate"
            meta["count"] = 3

        with tempfile.NamedTemporaryFile(mode="w") as temp_file:
            with rasterio.open(temp_file.name, "w", **meta) as dst:
                dst.write(data)

            s3_client = boto3.client("s3")
            s3_client.upload_file(temp_file.name, bucket, new_key)

        item = create_stac_item(new_href, with_proj=True)
        item.datetime = parser.parse(original_item.properties["start_datetime"])
        item.id = original_item.id

        # Convert Dictionary to JSON String
        data_string = json.dumps(item.to_dict())

        # Upload JSON String to an S3 Object
        s3 = boto3.resource("s3")
        s3_bucket = s3.Bucket(name=bucket)
        s3_bucket.put_object(
            Key=f"{PLATFORM_NAME}/{item.id}/stac_item.json",
            Body=data_string,
        )

        indexer = NoDataMaskChipIndexer(item, nodata_mask=data[0] == 0)
        index = indexer.create_index()

        writer = pa.BufferOutputStream()
        io.write_geoparquet_table(index, writer)
        body = bytes(writer.getvalue())
        # Centralize the index files to make combining them easier later on
        s3_bucket.put_object(
            Body=body,
            Key=f"index/{PLATFORM_NAME}/{item.id}/index_{item.id}.parquet",
        )


def process() -> None:

    if "AWS_BATCH_JOB_ARRAY_INDEX" not in os.environ:
        raise ValueError("AWS_BATCH_JOB_ARRAY_INDEX env var not set")
    if "STACCHIP_BUCKET" not in os.environ:
        raise ValueError("STACCHIP_BUCKET env var not set")

    index = int(os.environ["AWS_BATCH_JOB_ARRAY_INDEX"])
    bucket = os.environ["STACCHIP_BUCKET"]

    process_linz_tile(index, bucket)



import random

import pystac_client
from stacchip.indexer import NoStatsChipIndexer
from stacchip.chipper import Chipper
import os
import matplotlib.pyplot as plt

# Optimize GDAL settings for cloud optimized reading
os.environ["GDAL_DISABLE_READDIR_ON_OPEN"] = "EMPTY_DIR"
os.environ["AWS_REQUEST_PAYER"] = "requester"

# Query STAC catalog for NAIP data
catalog = pystac_client.Client.open("https://earth-search.aws.element84.com/v1")


items = catalog.search(
    collections=["naip"],
    max_items=1000,
)

items = items.item_collection()

items_list = list(items)
random.shuffle(items_list)

chips = []
for item in items_list[:10]:
    print(f"Working on {item}")

    # Index the chips in the item
    indexer = NoStatsChipIndexer(item)

    # Instanciate the chipper
    chipper = Chipper(indexer, asset_blacklist=["metadata"])

    # Get first chip for the "image" asset key
    for chip_id in random.sample(range(0, len(chipper)), 5):
        chips.append(chipper[chip_id]["image"])


fig, axs = plt.subplots(5, 10, gridspec_kw={'wspace': 0.01, 'hspace': 0.01}, squeeze=True)

for idx, ax in enumerate(axs.flatten()):
    chip = chips[idx]
    # Visualize the data
    ax.imshow(chip[:3].swapaxes(0, 1).swapaxes(1, 2))

plt.tight_layout()
plt.show()



with open("/home/tam/Desktop/all_naip_tiffs.txt") as src:
    paths = src.readlines()

from multiprocessing import Pool
from urllib.parse import urlparse

import rasterio


def drop(path):
    with rasterio.open(path) as rst:
        if rst.count != 4:
            url = urlparse(path)
            print(url)

            item_id = url.path.split("/")[2]
            print("Deleting", item_id)

            s3client = boto3.client("s3")
            s3client.delete_object(Bucket=url.netloc, Key=url.path.lstrip("/"))
            s3client.delete_object(
                Bucket=url.netloc, Key=f"naip/{item_id}/stac_item.json"
            )
            s3client.delete_object(
                Bucket=url.netloc, Key=f"index/naip/{item_id}/index_{item_id}.parquet"
            )
            # print(url.path.lstrip("/"))
            # print(f"index/naip/{item_id}/")
            # print(f"naip/{item_id}/{item_id}.parquet")
            return True


with Pool(24) as p:
    counts = p.map(drop, paths)



import json
import os
import random
from pathlib import Path
from urllib.parse import urlparse

import boto3
import geopandas as gp
import pyarrow as pa
import pystac_client
from geoarrow.pyarrow import io

from stacchip.indexer import LandsatIndexer

STAC_API = "https://landsatlook.usgs.gov/stac-server"

LS_ASSETS_L1 = [
    "blue",
    "green",
    "red",
    "nir08",
    "swir16",
    "lwir",
    "lwir_high",
    "swir22",
    "pan",
    "qa_pixel",
]
LS_ASSETS_L2 = [
    "blue",
    "green",
    "red",
    "nir08",
    "swir16",
    "swir22",
    "qa_pixel",
]
ABSOLUTE_CLOUD_COVER_FILTER = 75
PLATFORM_NAME_L2 = "landsat-c2l2-sr"
PLATFORM_NAME_L1 = "landsat-c2l1"
quartals = [
    "{year}-01-01/{year}-03-31",
    "{year}-04-01/{year}-06-30",
    "{year}-07-01/{year}-09-30",
    "{year}-10-01/{year}-12-31",
]


def process_landsat_tile(index: int, sample_source: str, bucket: str) -> None:
    # Prepare resources for the job
    catalog = pystac_client.Client.open(STAC_API)
    s3 = boto3.resource("s3")
    data = gp.read_file(sample_source)
    row = data.iloc[index]

    print("MGRS", row["name"])
    for platform_name in [PLATFORM_NAME_L1, PLATFORM_NAME_L2]:
        random.seed(index)
        for year in random.sample(range(2018, 2024), 1):
            print(f"Year {year}")
            for quartal in quartals:
                print(f"Quartal {quartal.format(year=year)}")
                items = catalog.search(
                    collections=[platform_name],
                    datetime=quartal.format(year=year),
                    max_items=1,
                    intersects=row.geometry.centroid,
                    sortby="properties.eo:cloud_cover",
                    query={
                        "platform": {"in": ["LANDSAT_8", "LANDSAT_9"]},
                    },
                )
                item = items.item_collection()[0]

                if item.properties["eo:cloud_cover"] > ABSOLUTE_CLOUD_COVER_FILTER:
                    continue

                print(
                    f"Cloud cover is {item.properties['eo:cloud_cover']} ({item.properties['platform']})"
                )

                for key in list(item.assets.keys()):
                    if (
                        platform_name == PLATFORM_NAME_L1 and key not in LS_ASSETS_L1
                    ) or (key not in LS_ASSETS_L2):
                        del item.assets[key]
                    else:
                        href = item.assets[key].extra_fields["alternate"]["s3"]["href"]
                        url = urlparse(href)
                        copy_source = {
                            "Bucket": url.netloc,
                            "Key": url.path.lstrip("/"),
                        }
                        print(f"Copying {key} band to {copy_source}")
                        new_key = f"{platform_name}/{item.id}/{Path(href).name}"
                        s3.meta.client.copy(
                            copy_source,
                            bucket,
                            new_key,
                            ExtraArgs={"RequestPayer": "requester"},
                        )
                        item.assets[key].href = f"s3://{bucket}/{new_key}"

                # Convert Dictionary to JSON String
                data_string = json.dumps(item.to_dict())

                # Upload JSON String to an S3 Object
                s3_bucket = s3.Bucket(name=bucket)
                s3_bucket.put_object(
                    Key=f"{platform_name}/{item.id}/stac_item.json",
                    Body=data_string,
                )

                indexer = LandsatIndexer(item, chip_max_nodata=0)
                chip_index = indexer.create_index()

                writer = pa.BufferOutputStream()
                io.write_geoparquet_table(chip_index, writer)
                body = bytes(writer.getvalue())
                # Centralize the index files to make combining them easier later on
                s3_bucket.put_object(
                    Body=body,
                    Key=f"index/{platform_name}/{item.id}/index_{item.id}.parquet",
                )


def process() -> None:

    if "AWS_BATCH_JOB_ARRAY_INDEX" not in os.environ:
        raise ValueError("AWS_BATCH_JOB_ARRAY_INDEX env var not set")
    if "STACCHIP_SAMPLE_SOURCE" not in os.environ:
        raise ValueError("STACCHIP_SAMPLE_SOURCE env var not set")
    if "STACCHIP_BUCKET" not in os.environ:
        raise ValueError("STACCHIP_BUCKET env var not set")

    index = int(os.environ["AWS_BATCH_JOB_ARRAY_INDEX"])
    sample_source = os.environ["STACCHIP_SAMPLE_SOURCE"]
    bucket = os.environ["STACCHIP_BUCKET"]

    process_landsat_tile(index, sample_source, bucket)



import math
from pathlib import Path
from typing import List, Optional
from urllib.parse import urlparse

import rasterio
from numpy.typing import ArrayLike
from rasterio.enums import Resampling
from rasterio.windows import Window

from stacchip.indexer import ChipIndexer


class Chipper:
    """
    Chipper class for managing and processing raster data chips.
    """

    def __init__(
        self,
        indexer: ChipIndexer,
        mountpath: Optional[str] = None,
        assets: Optional[List[str]] = None,
        asset_blacklist: Optional[List[str]] = None,
    ) -> None:
        """
        Initializes the Chipper class.

        Args:
            indexer (Type[ChipIndexer]): Input data which has to be of type ChipIndexer.
            mountpath (Optional[str]): Path to the mount directory for raster indexer.
                Defaults to None.
            assets (Optional[List[str]]): List of asset names to include for processing.
                If not provided, all assets are processed. Defaults to None.
            asset_blacklist (Optional[List[str]]): List of asset names to exclude from
                processing. Defaults to None.

        """
        self.mountpath = None if mountpath is None else Path(mountpath)
        self.assets = assets
        self.asset_blacklist = asset_blacklist
        self.indexer = indexer

    def __len__(self) -> int:
        """
        Returns the number of chips available.

        Returns:
            int: Number of chips available based on the indexer size.
        """
        return self.indexer.size

    def __getitem__(self, index: int) -> tuple:
        """
        Gets the chip by a single index.

        Args:
            index (int): Index of the chip to retrieve.

        Returns:
            tuple: A tuple containing x index, y index, and the chip data.
        """
        y_index = index // self.indexer.x_size
        x_index = index % self.indexer.x_size
        return x_index, y_index, self.chip(x_index, y_index)

    def __iter__(self):
        """
        Iterates over chips.

        Yields:
            tuple: The next chip data in the sequence.
        """
        counter = 0
        while counter < self.indexer.size:
            yield self[counter]
            counter += 1

    def get_pixels_for_asset(self, key: str, x: int, y: int) -> ArrayLike:
        """
        Extracts chip pixel values for one asset.

        Args:
            key (str): The asset key to extract pixels from.
            x (int): The x index of the chip.
            y (int): The y index of the chip.

        Returns:
            ArrayLike: Array of pixel values for the specified asset.

        Raises:
            ValueError: If asset dimensions are not multiples of the highest resolution dimensions.
        """
        asset = self.indexer.item.assets[key]

        srcpath = asset.href
        if self.mountpath:
            url = urlparse(srcpath, allow_fragments=False)
            srcpath = self.mountpath / Path(url.path.lstrip("/"))

        with rasterio.open(srcpath) as src:
            # Currently assume that different assets may be at different
            # resolutions, but are aligned and the gsd differs by an integer
            # multiplier.
            if self.indexer.shape[0] % src.height:
                raise ValueError(
                    f"Asset height {src.height} is not a multiple of highest resolution height {self.indexer.shape[0]}"  # noqa: E501
                )

            if self.indexer.shape[1] % src.width:
                raise ValueError(
                    f"Asset width {src.width} is not a multiple of highest resolution width {self.indexer.shape[1]}"  # noqa: E501
                )

            factor = self.indexer.shape[0] / src.height

            chip_window = Window(
                math.floor(x * self.indexer.chip_size / factor),
                math.floor(y * self.indexer.chip_size / factor),
                math.ceil(self.indexer.chip_size / factor),
                math.ceil(self.indexer.chip_size / factor),
            )

            return src.read(
                window=chip_window,
                out_shape=(src.count, self.indexer.chip_size, self.indexer.chip_size),
                resampling=Resampling.nearest,
            )

    def chip(self, x: int, y: int) -> dict:
        """
        Retrieves chip pixel array for the specified x and y index numbers.

        Args:
            x (int): The x index of the chip.
            y (int): The y index of the chip.

        Returns:
            dict: A dictionary where keys are asset names and values are arrays of pixel values.
        """
        if self.assets is not None:
            keys = self.assets
        else:
            keys = list(self.indexer.item.assets.keys())

        if self.asset_blacklist is not None:
            keys = [key for key in keys if key not in self.asset_blacklist]

        return {key: self.get_pixels_for_asset(key, x, y) for key in keys}



from multiprocessing import Pool
from pathlib import Path

import numpy as np

src = Path("/home/tam/Desktop/clay-v1-data-chips-naip")

paths = [path for path in src.glob("**/*.npz")]


def process(path):
    print(path)
    data = dict(np.load(path))
    data["pixels"] = data["pixels"][:, :224, :224]
    counter = path.stem.split("_")[1]
    np.savez_compressed(
        f"/home/tam/Desktop/clay-v1-data-chips-naip-224/chip_{counter}.npz",
        data["pixels"],
        data["lon_norm"],
        data["lat_norm"],
        data["week_norm"],
        data["hour_norm"],
    )


with Pool(12) as p:
    p.map(process, paths)



import json
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
import rasterio
from numpy.testing import assert_array_equal
from pystac import Item

from stacchip.chipper import Chipper
from stacchip.indexer import NoStatsChipIndexer


def test_no_stats_indexer():
    with TemporaryDirectory() as dirname:
        mountpath = Path(dirname)
        target_dir = mountpath / "naip/item1"
        target_dir.mkdir(parents=True, exist_ok=True)
        item = Item.from_file("tests/data/stacchip_test_item.json")
        shape = item.properties["proj:shape"]
        size = shape[0] * shape[1]
        trsf = item.properties["proj:transform"]
        bands = 2
        with rasterio.open(
            mountpath / "naip/item1/asset.tif",
            "w",
            width=shape[1],
            height=shape[0],
            count=bands,
            dtype="uint8",
            transform=[trsf[2], trsf[0], trsf[1], trsf[5], trsf[4], trsf[3]],
        ) as rst:
            raster_data = np.random.randint(
                0, 255, bands * size, dtype="uint8"
            ).reshape((bands, *shape))
            rst.write(raster_data)

        item.assets["asset"].href = "s3://example-bucket/naip/item1/asset.tif"
        with open(mountpath / "naip/item1/stac_item.json", "w") as dst:
            dst.write(json.dumps(item.to_dict()))
        indexer = NoStatsChipIndexer(item)
        index = indexer.create_index()
        chipper = Chipper(indexer, mountpath=mountpath)
        x = index.column("chip_index_x")[1].as_py()
        y = index.column("chip_index_y")[2].as_py()
        chip = chipper.chip(x, y)
        assert chip["asset"].shape[0] == raster_data.shape[0]
        assert_array_equal(
            chip["asset"][0],
            raster_data[
                0,
                (y * indexer.chip_size) : ((y + 1) * indexer.chip_size),
                (x * indexer.chip_size) : ((x + 1) * indexer.chip_size),
            ],
        )
        # Test magic functions
        assert len(chipper) == indexer.size
        x_index, y_index, chipper_1 = chipper[1]
        assert x == x_index
        assert y == y_index
        assert_array_equal(chip["asset"][0], chipper_1["asset"][0])
        counter = 0
        for _chip in chipper:
            counter += 1
        assert counter == len(chipper)



import warnings
from functools import cached_property
from math import floor
from typing import Tuple

import geoarrow.pyarrow as ga
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import pyproj
import rasterio
from numpy.typing import ArrayLike
from pystac import Item
from rasterio.crs import CRS
from rasterio.enums import Resampling
from shapely import GeometryType
from shapely.geometry import box
from shapely.ops import transform

warnings.filterwarnings(
    "ignore",
    message=(
        "The argument 'infer_datetime_format' is deprecated and will"
        " be removed in a future version. A strict version of it is now "
        "the default, see https://pandas.pydata.org/pdeps/0004-consistent"
        "-to-datetime-parsing.html. You can safely remove this argument."
    ),
)


class ChipIndexer:
    """
    Indexer base class
    """

    def __init__(
        self, item: Item, chip_size: int = 256, chip_max_nodata: float = 0.5
    ) -> None:
        """
        Init ChipIndexer
        """
        self.item = item
        self.chip_size = chip_size
        self.chip_max_nodata = chip_max_nodata

        assert self.item.ext.has("proj")

        self.assert_units_metre()
        self.setup_projector()

    def assert_units_metre(self) -> None:
        """
        Ensure input data has meters as units
        """
        crs = CRS(init=f"EPSG:{self.item.properties['proj:epsg']}")
        assert crs.linear_units == "metre"

    def setup_projector(self):
        """
        Prepare projection function to project geometries into WGS84
        """
        wgs84 = pyproj.CRS("EPSG:4326")
        target_crs = pyproj.CRS(f'EPSG:{self.item.properties["proj:epsg"]}')
        self._projector = pyproj.Transformer.from_crs(
            target_crs, wgs84, always_xy=True
        ).transform

    def reproject(self, geom) -> GeometryType:
        """
        Reproject a geometry into WGS84
        """
        return transform(self._projector, geom)

    def _get_trsf_or_shape(self, key: str) -> list:
        """
        The shape of the hightest resolution band
        """
        data = []
        if key in self.item.properties:
            data = self.item.properties[key]
        else:
            for asset in self.item.assets.values():
                if key not in asset.extra_fields:
                    continue
                if not data or data[0] < asset.extra_fields[key][0]:
                    data = asset.extra_fields[key]

        return data

    @cached_property
    def shape(self) -> list:
        """
        Shape of the STAC item data

        Obtains the shape of the highest resolution band from
        all the available bands.
        """
        return self._get_trsf_or_shape("proj:shape")

    @cached_property
    def transform(self) -> list:
        """
        The transform property from the STAC item
        """
        return self._get_trsf_or_shape("proj:transform")

    @property
    def x_size(self) -> int:
        """
        Number of tiles vailable in x direction
        """
        return floor(self.shape[1] / self.chip_size)

    @property
    def y_size(self) -> int:
        """
        Number of tiles vailable in y direction
        """
        return floor(self.shape[0] / self.chip_size)

    @property
    def size(self) -> int:
        """
        Number of tiles in this STAC item
        """
        return self.x_size * self.y_size

    @property
    def bbox(self) -> Tuple[float, float, float, float]:
        """
        Bounding box that covers all tiles

        This is different from the bounding box of the STAC item
        if the tiles don't fit into the number of pixels perfectly.
        """
        return (
            self.transform[2],
            self.transform[5] + self.transform[4] * self.shape[0],
            self.transform[2] + self.transform[0] * self.shape[1],
            self.transform[5],
        )

    def get_stats(self, x: int, y: int) -> Tuple[float, float]:
        """
        A function to write for each indexer that returns nodata and
        cloud statistics for a chip
        """
        raise NotImplementedError()

    def get_chip_bbox(self, x: int, y: int) -> str:
        """
        Bounding box for a chip
        """
        chip_box = box(
            self.bbox[0] + x * self.transform[0] * self.chip_size,
            self.bbox[3] + y * self.transform[4] * self.chip_size,
            self.bbox[0] + (x + 1) * self.transform[0] * self.chip_size,
            self.bbox[3] + (y + 1) * self.transform[4] * self.chip_size,
        )

        return self.reproject(chip_box).wkt

    def create_index(self) -> pa.Table:
        """
        The index for this STAC item
        """
        index = {
            "chipid": np.empty(self.size, dtype="<U256"),
            "date": np.empty(self.size, dtype="datetime64[D]"),
            "chip_index_x": np.empty(self.size, dtype="uint16"),
            "chip_index_y": np.empty(self.size, dtype="uint16"),
            "cloud_cover_percentage": np.empty(self.size, dtype="float32"),
            "nodata_percentage": np.empty(self.size, dtype="float32"),
            "geometry": np.empty(self.size, dtype="object"),
        }
        counter = 0
        for y in range(0, self.y_size):
            for x in range(0, self.x_size):

                cloud_cover_percentage, nodata_percentage = self.get_stats(x, y)

                index["chipid"][counter] = f"{self.item.id}-{x}-{y}"
                index["date"][counter] = self.item.datetime.date()
                index["chip_index_x"][counter] = x
                index["chip_index_y"][counter] = y
                index["cloud_cover_percentage"][counter] = cloud_cover_percentage
                index["nodata_percentage"][counter] = nodata_percentage
                index["geometry"][counter] = self.get_chip_bbox(x, y)

                counter += 1

        index["geometry"] = ga.as_geoarrow(index["geometry"])

        table = pa.table(index)
        chips_count = table.shape[0]
        table = table.filter(pc.field("nodata_percentage") <= self.chip_max_nodata)
        print(
            f"Dropped {chips_count - table.shape[0]}/{chips_count} chips due to nodata above {self.chip_max_nodata}"
        )
        return table


class NoStatsChipIndexer(ChipIndexer):
    """
    Indexer that assumes that none of the chips have any clouds or nodata
    """

    def get_stats(self, x: int, y: int) -> Tuple[float, float]:
        """
        Cloud and nodata percentage for a chip
        """
        return 0.0, 0.0


class NoDataMaskChipIndexer(ChipIndexer):
    """
    Chip indexer that takes the nodata mask as input and assumes that
    there are no clouds in the image
    """

    def __init__(
        self,
        item: Item,
        nodata_mask: ArrayLike,
        chip_size: int = 256,
        chip_max_nodata: float = 0.5,
    ) -> None:
        """
        Init NoDataMaskChipIndexer
        """
        super().__init__(item, chip_size, chip_max_nodata)
        self.nodata_mask = nodata_mask

    def get_stats(self, x: int, y: int) -> Tuple[float, float]:
        """
        Cloud and nodata percentage for a chip

        Assumes there are no cloudy pixels and computes nodata from mask
        """
        nodata_percentage = np.sum(
            self.nodata_mask[
                y * self.chip_size : (y + 1) * self.chip_size,
                x * self.chip_size : (x + 1) * self.chip_size,
            ]
        ) / (self.chip_size**2)

        return 0.0, nodata_percentage


class LandsatIndexer(ChipIndexer):
    """
    Chip indexer for Landsat 8 and 9 STAC items
    """

    @cached_property
    def qa(self):
        """
        The quality band data for the STAC item
        """
        print("Loading qa band")
        self.item.assets["qa_pixel"].href = self.item.assets["qa_pixel"].extra_fields[
            "alternate"
        ]["s3"]["href"]
        with rasterio.open(self.item.assets["qa_pixel"].href) as src:
            return src.read(1)

    def get_stats(self, x: int, y: int) -> Tuple[float, float]:
        """
        Cloud and nodata percentage for a chip

        Uses the qa band to compute these values.
        """
        qa = self.qa[
            y * self.chip_size : (y + 1) * self.chip_size,
            x * self.chip_size : (x + 1) * self.chip_size,
        ]

        # Bit 1 is dilated cloud, 3 is cloud, 4 is cloud shadow.
        nodata_byte = np.array(1 << 0, dtype=qa.dtype)
        dilated_cloud_byte = np.array(1 << 1, dtype=qa.dtype)
        cloud_byte = np.array(1 << 3, dtype=qa.dtype)
        shadow_byte = np.array(1 << 4, dtype=qa.dtype)

        nodata_mask = np.bitwise_and(qa, nodata_byte)
        dilated_cloud = np.bitwise_and(qa, dilated_cloud_byte)
        cloud = np.bitwise_and(qa, cloud_byte)
        shadow = np.bitwise_and(qa, shadow_byte)

        layer_clouds = (dilated_cloud | cloud | shadow).astype(dtype="bool")

        cloud_percentage = np.sum(layer_clouds) / qa.size
        nodata_percentage = np.sum(nodata_mask) / qa.size

        return cloud_percentage, nodata_percentage


class Sentinel2Indexer(ChipIndexer):
    """
    Indexer for Sentinel-2 STAC items
    """

    scl_filter = [1, 3, 8, 9, 10]
    nodata_value = 0

    @cached_property
    def scl(self):
        """
        The Scene Classification (SCL) band data for the STAC item
        """
        print("Loading scl band")
        with rasterio.open(self.item.assets["scl"].href) as src:
            return src.read(out_shape=(1, *self.shape), resampling=Resampling.nearest)[
                0
            ]

    def get_stats(self, x: int, y: int) -> Tuple[float, float]:
        """
        Cloud and nodata percentage for a chip

        Uses the SCL band to compute these values.
        """
        scl = self.scl[
            y * self.chip_size : (y + 1) * self.chip_size,
            x * self.chip_size : (x + 1) * self.chip_size,
        ]

        cloud_percentage = int(np.isin(scl, self.scl_filter).sum()) / scl.size

        nodata_percentage = np.sum(scl == self.nodata_value) / scl.size

        return cloud_percentage, nodata_percentage



import json
import os
import random
from pathlib import Path
from urllib.parse import urlparse

import boto3
import geopandas as gp
import pyarrow as pa
import pystac_client
from geoarrow.pyarrow import io

from stacchip.indexer import Sentinel2Indexer

STAC_API = "https://earth-search.aws.element84.com/v1"
S2_ASSETS = [
    "blue",
    "green",
    "nir",
    "nir08",
    "red",
    "rededge1",
    "rededge2",
    "rededge3",
    "scl",
    "swir16",
    "swir22",
]
ABSOLUTE_CLOUD_COVER_FILTER = 0.75
PLATFORM_NAME = "sentinel-2-l2a"
SCENE_NODATA_LIMIT = 20
quartals = [
    "{year}-01-01/{year}-03-31",
    "{year}-04-01/{year}-06-30",
    "{year}-07-01/{year}-09-30",
    "{year}-10-01/{year}-12-31",
]


def process_mgrs_tile(index: int, mgrs_source: str, bucket: str) -> None:
    # Prepare resources for the job
    catalog = pystac_client.Client.open(STAC_API)
    s3 = boto3.resource("s3")
    data = gp.read_file(mgrs_source)
    row = data.iloc[index]

    print("MGRS", row["name"])
    random.seed(index)
    for year in random.sample(range(2018, 2024), 2):
        print(f"Year {year}")
        for quartal in quartals:
            print(f"Quartal {quartal.format(year=year)}")
            items = catalog.search(
                collections=["sentinel-2-l2a"],
                datetime=quartal.format(year=year),
                max_items=1,
                intersects=row.geometry,
                sortby="properties.eo:cloud_cover",
                query={
                    "grid:code": {
                        "eq": f"MGRS-{row['name']}",
                    },
                    "s2:nodata_pixel_percentage": {"lte": SCENE_NODATA_LIMIT},
                },
            )
            item = items.item_collection()[0]

            if item.properties["eo:cloud_cover"] > ABSOLUTE_CLOUD_COVER_FILTER:
                continue

            print(f"Cloud cover is {item.properties['eo:cloud_cover']}")

            for key in list(item.assets.keys()):
                if key not in S2_ASSETS:
                    del item.assets[key]
                else:
                    url = urlparse(item.assets[key].href)
                    copy_source = {
                        "Bucket": "sentinel-cogs",
                        "Key": url.path.lstrip("/"),
                    }
                    print(f"Copying {copy_source}")
                    new_key = (
                        f"{PLATFORM_NAME}/{item.id}/{Path(item.assets[key].href).name}"
                    )
                    s3.meta.client.copy(copy_source, bucket, new_key)
                    item.assets[key].href = f"s3://{bucket}/{new_key}"

            # Convert Dictionary to JSON String
            data_string = json.dumps(item.to_dict())

            # Upload JSON String to an S3 Object
            s3_bucket = s3.Bucket(name=bucket)
            s3_bucket.put_object(
                Key=f"{PLATFORM_NAME}/{item.id}/stac_item.json",
                Body=data_string,
            )

            indexer = Sentinel2Indexer(item)
            index = indexer.create_index()

            writer = pa.BufferOutputStream()
            io.write_geoparquet_table(index, writer)
            body = bytes(writer.getvalue())
            # Centralize the index files to make combining them easier later on
            s3_bucket.put_object(
                Body=body,
                Key=f"index/{PLATFORM_NAME}/{item.id}/index_{item.id}.parquet",
            )


def process() -> None:

    if "AWS_BATCH_JOB_ARRAY_INDEX" not in os.environ:
        raise ValueError("AWS_BATCH_JOB_ARRAY_INDEX env var not set")
    if "STACCHIP_MGRS_SOURCE" not in os.environ:
        raise ValueError("STACCHIP_MGRS_SOURCE env var not set")
    if "STACCHIP_BUCKET" not in os.environ:
        raise ValueError("STACCHIP_BUCKET env var not set")

    index = int(os.environ["AWS_BATCH_JOB_ARRAY_INDEX"])
    mgrs_source = os.environ["STACCHIP_MGRS_SOURCE"]
    bucket = os.environ["STACCHIP_BUCKET"]

    process_mgrs_tile(index, mgrs_source, bucket)



import datetime
import math
import os
from io import BytesIO
from multiprocessing import Pool
from typing import Union

import boto3
import numpy as np
import pyarrow as pa
import shapely
from pyarrow import dataset as da

from stacchip.chipper import Chipper

VERSION = "mode_v1_chipper_v2"

CUBESIZE = 128

S2_BANDS = [
    "blue",
    "green",
    "red",
    "rededge1",
    "rededge2",
    "rededge3",
    "nir",
    "nir08",
    "swir16",
    "swir22",
]
LS_BANDS = [
    "red",
    "green",
    "blue",
    "nir08",
    "swir16",
    "swir22",
]
NAIP_BANDS = ["red", "green", "blue", "nir"]
LINZ_BANDS = ["red", "green", "blue"]
S1_BANDS = ["vv", "vh"]


def normalize_timestamp(date):

    week = date.isocalendar().week * 2 * np.pi / 52
    hour = date.hour * 2 * np.pi / 24

    return (math.sin(week), math.cos(week)), (math.sin(hour), math.cos(hour))


def normalize_latlon(bounds):
    bounds = shapely.from_wkt(bounds).bounds
    lon = bounds[0] + (bounds[2] - bounds[0]) / 2
    lat = bounds[1] + (bounds[3] - bounds[1]) / 2

    lat = lat * np.pi / 180
    lon = lon * np.pi / 180

    return (math.sin(lat), math.cos(lat)), (math.sin(lon), math.cos(lon))


def stack_chips(chips: list, cube_id: int, chip_bucket: str, platform: str):
    print(f"Writing cube {cube_id}")

    pixels = np.stack([chip["pixels"] for chip in chips], dtype="float32")
    lon_norm = np.vstack([chip["lon_norm"] for chip in chips], dtype="float32")
    lat_norm = np.vstack([chip["lat_norm"] for chip in chips], dtype="float32")
    week_norm = np.vstack([chip["week_norm"] for chip in chips], dtype="float32")
    hour_norm = np.vstack([chip["hour_norm"] for chip in chips], dtype="float32")

    key = f"{VERSION}/{platform}/cube_{cube_id}.npz"

    client = boto3.client("s3")
    with BytesIO() as bytes:
        np.savez_compressed(
            file=bytes,
            pixels=pixels,
            lon_norm=lon_norm,
            lat_norm=lat_norm,
            week_norm=week_norm,
            hour_norm=hour_norm,
        )
        bytes.seek(0)
        client.upload_fileobj(Fileobj=bytes, Bucket=chip_bucket, Key=key)


def get_chip(
    data_bucket: str,
    row: int,
    platform: str,
    item_id: str,
    date: Union[datetime.date, datetime.datetime],
    chip_index_x: int,
    chip_index_y: int,
):
    print(
        "Getting chip",
        data_bucket,
        row,
        platform,
        item_id,
        date,
        chip_index_x,
        chip_index_y,
    )

    chipper = Chipper(
        bucket=data_bucket,
        platform=platform,
        item_id=item_id,
    )
    chip = chipper.chip(chip_index_x, chip_index_y)

    if platform == "naip":
        pixels = chip["image"]
        bands = NAIP_BANDS
    elif platform == "linz":
        pixels = chip["asset"]
        bands = LINZ_BANDS
    elif platform == "sentinel-2-l2a":
        pixels = np.vstack([chip[band] for band in S2_BANDS])
        bands = S2_BANDS
    elif platform in ["landsat-c2l2-sr", "landsat-c2l1"]:
        pixels = np.vstack([chip[band] for band in LS_BANDS])
        bands = LS_BANDS
    elif platform == "sentinel-1-rtc":
        if any(band not in chip for band in S1_BANDS):
            return
        pixels = np.vstack([chip[band] for band in S1_BANDS])
        bands = S1_BANDS

    if len(pixels) != len(bands):
        raise ValueError(
            f"Pixels shape {pixels.shape} is not equal to nr of bands {bands} for item {item_id}"
        )

    if isinstance(date, datetime.date):
        # Assume noon for dates without timestamp
        date = datetime.datetime(date.year, date.month, date.day, 12)
    week_norm, hour_norm = normalize_timestamp(date)

    bounds = chipper.indexer.get_chip_bbox(chip_index_x, chip_index_y)
    lon_norm, lat_norm = normalize_latlon(bounds)

    return {
        "pixels": pixels,
        "lon_norm": lon_norm,
        "lat_norm": lat_norm,
        "week_norm": week_norm,
        "hour_norm": hour_norm,
    }


def process() -> None:
    # GDAL read optimization is recommended
    # os.environ["GDAL_DISABLE_READDIR_ON_OPEN"] = "YES"
    # os.environ["CPL_VSIL_CURL_ALLOWED_EXTENSIONS"] = ".tif,.png,.jp2,.tiff"

    if "AWS_BATCH_JOB_ARRAY_INDEX" not in os.environ:
        raise ValueError("AWS_BATCH_JOB_ARRAY_INDEX env var not set")
    if "STACCHIP_DATA_BUCKET" not in os.environ:
        raise ValueError("STACCHIP_DATA_BUCKET env var not set")
    if "STACCHIP_INDEXPATH" not in os.environ:
        raise ValueError("STACCHIP_INDEXPATH env var not set")
    if "STACCHIP_CHIP_BUCKET" not in os.environ:
        raise ValueError("STACCHIP_TARGETPATH env var not set")

    index = int(os.environ["AWS_BATCH_JOB_ARRAY_INDEX"])
    data_bucket = os.environ["STACCHIP_DATA_BUCKET"]
    indexpath = os.environ["STACCHIP_INDEXPATH"]
    chip_bucket = os.environ["STACCHIP_CHIP_BUCKET"]
    platform = os.environ.get("STACCHIP_PLATFORM")
    cubes_per_job = int(os.environ.get("STACCHIP_CUBES_PER_JOB", 10))
    pool_size = int(os.environ.get("STACCHIP_POOL_SIZE", 10))
    chip_max_nodata = int(os.environ.get("STACCHIP_MAX_NODATA", 0.05))

    # Open table
    table = da.dataset(indexpath, format="parquet").to_table(
        columns=["chipid", "platform", "item", "date", "chip_index_x", "chip_index_y"]
    )
    if platform:
        table = table.filter(pa.compute.field("platform") == platform)

    initial_count = len(table)
    if chip_max_nodata:
        table = table.filter(pa.compute.field("nodata_percentage") <= chip_max_nodata)
    print(
        f"Dropped {initial_count - len(table)} chips due to nodata filter, keeping {len(table)}"
    )

    np.random.seed(42)
    random_rows = np.random.randint(0, len(table), len(table))

    for cube_id in range(index * cubes_per_job, (index + 1) * cubes_per_job):
        random_rows_cube = random_rows[cube_id * CUBESIZE : (cube_id + 1) * CUBESIZE]
        if len(random_rows_cube) != CUBESIZE:
            print("Finishing because of incomplete cubes")
            return

        # Extract chips data for this job
        all_chips = []
        for row in random_rows_cube:
            all_chips.append(
                (
                    data_bucket,
                    row,
                    table.column("platform")[row].as_py(),
                    table.column("item")[row].as_py(),
                    table.column("date")[row].as_py(),
                    table.column("chip_index_x")[row].as_py(),
                    table.column("chip_index_y")[row].as_py(),
                )
            )

        with Pool(pool_size) as pl:
            data = pl.starmap(
                get_chip,
                all_chips,
            )

        if None in data:
            print(f"Not all cubes are complete, skipping stacking for cube {cube_id}")
            continue

        stack_chips(data, cube_id=cube_id, chip_bucket=chip_bucket, platform=platform)



import datetime

import mock
import numpy as np
import pyarrow as pa
import pytest
from pystac import Item
from rasterio import Affine
from rasterio.io import MemoryFile
from shapely import Point

from stacchip.indexer import (
    ChipIndexer,
    LandsatIndexer,
    NoStatsChipIndexer,
    Sentinel2Indexer,
)


def get_ls_mock(nodata: bool = False) -> MemoryFile:
    meta = {
        "driver": "GTiff",
        "dtype": "uint16",
        "nodata": None,
        "width": 8331,
        "height": 8271,
        "count": 1,
        "crs": "EPSG:3031",
        "transform": Affine(30.0, 0.0, 1517085.0, 0.0, -30.0, -1811685.0),
    }
    data = np.zeros((1, 8331, 8271))
    if nodata:
        data[0, :200, :200] = 1
    memfile = MemoryFile()
    with memfile.open(**meta) as dst:
        dst.write(data)
    return memfile.open()


def rasterio_open_ls_mock(href: str) -> MemoryFile:
    return get_ls_mock()


def rasterio_open_ls_nodata_mock(href: str) -> MemoryFile:
    return get_ls_mock(True)


def rasterio_open_sentinel_mock(href: str) -> MemoryFile:
    meta = {
        "driver": "GTiff",
        "dtype": "uint8",
        "nodata": 0.0,
        "width": 5490,
        "height": 5490,
        "count": 1,
        "crs": "EPSG:32720",
        "transform": Affine(20.0, 0.0, 499980.0, 0.0, -20.0, 6400000.0),
    }
    data = 5 * np.ones((1, 5490, 5490), dtype="uint16")
    # Make first chip nodata
    data[0, :100, :100] = 0
    # Make second chip cloudy
    data[0, :128, 128:192] = 1
    memfile = MemoryFile()
    with memfile.open(**meta) as dst:
        dst.write(data)
    return memfile.open()


def test_get_stats_error():
    item = Item.from_file("tests/data/naip_m_4207009_ne_19_060_20211024.json")
    indexer = ChipIndexer(item)
    with pytest.raises(NotImplementedError):
        indexer.create_index()


def test_no_stats_indexer():
    item = Item.from_file("tests/data/naip_m_4207009_ne_19_060_20211024.json")
    indexer = NoStatsChipIndexer(item)
    assert indexer.shape == [12666, 9704]
    index = indexer.create_index()
    assert str(index.column("chipid")[0]) == "m_4207009_ne_19_060_20211024.tif-0-0"
    assert index.column("date")[0] == pa.scalar(
        datetime.date(2021, 10, 24), pa.date32()
    )

    point = Point(
        indexer.transform[2],
        indexer.transform[5],
    )

    target = indexer.reproject(point)
    assert min(
        [dat["x"] for dat in index.column("geometry")[0].as_py()[0]]
    ) == pytest.approx(target.bounds[0])
    assert max(
        [dat["y"] for dat in index.column("geometry")[0].as_py()[0]]
    ) == pytest.approx(target.bounds[3])


@mock.patch("stacchip.indexer.rasterio.open", rasterio_open_sentinel_mock)
def test_sentinel_2_indexer():
    item = Item.from_file(
        "tests/data/sentinel-2-l2a-S2A_T20HNJ_20240311T140636_L2A.json"
    )
    indexer = Sentinel2Indexer(item)
    assert indexer.shape == [10980, 10980]
    index = indexer.create_index()
    assert index.shape == (1763, 7)
    assert str(index.column("chipid")[0]) == "S2A_T20HNJ_20240311T140636_L2A-1-0"
    assert index.column("cloud_cover_percentage")[0].as_py() == 0.5


@mock.patch("stacchip.indexer.rasterio.open", rasterio_open_ls_mock)
def test_landsat_indexer():
    item = Item.from_file(
        "tests/data/landsat-c2l2-sr-LC09_L2SR_086107_20240311_20240312_02_T2_SR.json"
    )
    indexer = LandsatIndexer(item)
    assert indexer.shape == [8271, 8331]
    index = indexer.create_index()
    assert isinstance(index, pa.Table)
    assert (
        str(index.column("chipid")[0])
        == "LC09_L2SR_086107_20240311_20240312_02_T2_SR-0-0"
    )
    assert index.shape == (1024, 7)
    assert indexer.x_size == int(8331 / 256)
    assert indexer.y_size == int(8271 / 256)


@mock.patch("stacchip.indexer.rasterio.open", rasterio_open_ls_nodata_mock)
def test_landsat_indexer_nodata():
    item = Item.from_file(
        "tests/data/landsat-c2l2-sr-LC09_L2SR_086107_20240311_20240312_02_T2_SR.json"
    )
    indexer = LandsatIndexer(item)
    index = indexer.create_index()
    assert index.shape == (1023, 7)
    assert (
        str(index.column("chipid")[0])
        == "LC09_L2SR_086107_20240311_20240312_02_T2_SR-1-0"
    )

    indexer = LandsatIndexer(item, chip_max_nodata=0.95)
    index = indexer.create_index()
    assert index.shape == (1024, 7)



import json
import os
import random
from pathlib import Path
from urllib.parse import urlparse

import boto3
import geopandas as gp
import pyarrow as pa
import pystac_client
from botocore.exceptions import ClientError
from geoarrow.pyarrow import io

from stacchip.indexer import NoStatsChipIndexer

STAC_API = "https://planetarycomputer.microsoft.com/api/stac/v1"

AWS_S3_URL = (
    "s3://naip-analytic/{state}/{year}/{resolution}/rgbir_cog/{block}{subblock}/{name}"
)
PLATFORM_NAME = "naip"


def process_naip_tile(
    index: int, sample_source: str, bucket: str, latest_only: bool = False
) -> None:
    # Prepare resources for the job
    catalog = pystac_client.Client.open(STAC_API)
    s3 = boto3.resource("s3")
    data = gp.read_file(sample_source)
    row = data.iloc[index]

    items = catalog.search(
        collections=["naip"],
        intersects=row.geometry.centroid,
        sortby="properties.naip:year",
        max_items=10,
    )
    items = list(items.item_collection())

    if not len(items):
        print(f"No items found, skipping index {index}")
        return

    latest_item = items.pop()
    items_to_process = [latest_item]
    if not latest_only:
        random.seed(index)
        random_item = random.choice(items)
        items_to_process.append(random_item)

    for item in items_to_process:
        print(f"Processing item {item.id}")
        for key in list(item.assets.keys()):
            if key != "image":
                del item.assets[key]
                continue

            new_key = f"{PLATFORM_NAME}/{item.id}/{Path(item.assets[key].href).name}"
            try:
                href = AWS_S3_URL.format(
                    year=item.properties["naip:year"],
                    state=item.properties["naip:state"],
                    resolution=f"{int(item.properties['gsd'] * 100)}cm",
                    block=item.id.split("_")[2][:5],
                    subblock=f"/{item.id.split('_')[2][5:]}",
                    name=item.assets["image"].href.split("/")[-1],
                )
                url = urlparse(href)
                copy_source = {
                    "Bucket": "naip-analytic",
                    "Key": url.path.lstrip("/"),
                }
                print(f"Copying {copy_source}")
                s3.Object("naip-analytic", url.path.lstrip("/")).load(
                    RequestPayer="requester"
                )
                s3.meta.client.copy(
                    copy_source,
                    bucket,
                    new_key,
                    ExtraArgs={"RequestPayer": "requester"},
                )
            except ClientError:
                href = AWS_S3_URL.format(
                    year=item.properties["naip:year"],
                    state=item.properties["naip:state"],
                    resolution=f"{int(item.properties['gsd'] * 100)}cm",
                    block=item.id.split("_")[2][:5],
                    subblock="",
                    name=item.assets["image"].href.split("/")[-1],
                )
                url = urlparse(href)
                copy_source = {
                    "Bucket": "naip-analytic",
                    "Key": url.path.lstrip("/"),
                }
                print(f"Failed, now copying {copy_source}")
                s3.Object("naip-analytic", url.path.lstrip("/")).load(
                    RequestPayer="requester"
                )
                s3.meta.client.copy(
                    copy_source,
                    bucket,
                    new_key,
                    ExtraArgs={"RequestPayer": "requester"},
                )

            item.assets[key].href = f"s3://{bucket}/{new_key}"

        # Convert Dictionary to JSON String
        data_string = json.dumps(item.to_dict())

        # Upload JSON String to an S3 Object
        s3_bucket = s3.Bucket(name=bucket)
        s3_bucket.put_object(
            Key=f"{PLATFORM_NAME}/{item.id}/stac_item.json",
            Body=data_string,
        )

        indexer = NoStatsChipIndexer(item)
        index = indexer.create_index()
        print("Indexer info", indexer.x_size, indexer.y_size, indexer.shape)

        writer = pa.BufferOutputStream()
        io.write_geoparquet_table(index, writer)
        body = bytes(writer.getvalue())
        # Centralize the index files to make combining them easier later on
        s3_bucket.put_object(
            Body=body,
            Key=f"index/{PLATFORM_NAME}/{item.id}/index_{item.id}.parquet",
        )


def process() -> None:

    if "AWS_BATCH_JOB_ARRAY_INDEX" not in os.environ:
        raise ValueError("AWS_BATCH_JOB_ARRAY_INDEX env var not set")
    if "STACCHIP_SAMPLE_SOURCE" not in os.environ:
        raise ValueError("STACCHIP_SAMPLE_SOURCE env var not set")
    if "STACCHIP_BUCKET" not in os.environ:
        raise ValueError("STACCHIP_BUCKET env var not set")

    index = int(os.environ["AWS_BATCH_JOB_ARRAY_INDEX"])
    sample_source = os.environ["STACCHIP_SAMPLE_SOURCE"]
    bucket = os.environ["STACCHIP_BUCKET"]

    process_naip_tile(index, sample_source, bucket)
